import os

import pytest

import salt.modules.environ as envmodule
import salt.modules.reg
import salt.states.environ as envstate
import salt.utils.platform
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    loader_globals = {
        "__env__": "base",
        "__opts__": {"test": False},
        "__salt__": {"environ.setenv": envmodule.setenv},
    }
    return {envstate: loader_globals, envmodule: loader_globals}


def test_setenv():
    """
    test that a subsequent calls of setenv changes nothing
    """
    ret = envstate.setenv("test", "value")
    assert ret["changes"] == {"test": "value"}

    ret = envstate.setenv("test", "other")
    assert ret["changes"] == {"test": "other"}

    # once again with the same value
    ret = envstate.setenv("test", "other")
    assert ret["changes"] == {}


@pytest.mark.skip_unless_on_windows
def test_setenv_permanent():
    """
    test that we can set permanent environment variables (requires pywin32)
    """
    with patch.dict(
        envmodule.__utils__,
        {
            "reg.set_value": MagicMock(),
            "reg.delete_value": MagicMock(),
            "win_functions.broadcast_setting_change": MagicMock(),
        },
    ):
        ret = envstate.setenv("test", "value", permanent=True)
        assert ret["changes"] == {"test": "value"}
        envmodule.__utils__["reg.set_value"].assert_called_with(
            "HKCU", "Environment", "test", "value"
        )

        ret = envstate.setenv("test", False, false_unsets=True, permanent=True)
        assert ret["changes"] == {"test": None}
        envmodule.__utils__["reg.delete_value"].assert_called_with(
            "HKCU", "Environment", "test"
        )


def test_setenv_dict():
    """
    test that setenv can be invoked with dict
    """
    ret = envstate.setenv("notimportant", {"test": "value"})
    assert ret["changes"] == {"test": "value"}


def test_setenv_int():
    """
    test that setenv can not be invoked with int
    (actually it's anything other than strings and dict)
    """
    ret = envstate.setenv("test", 1)
    assert ret["result"] is False


def test_setenv_unset():
    """
    test that ``false_unsets`` option removes variable from environment
    """
    with patch.dict(os.environ, {"INITIAL": "initial"}, clear=True):
        ret = envstate.setenv("test", "value")
        assert ret["changes"] == {"test": "value"}

        ret = envstate.setenv("notimportant", {"test": False}, false_unsets=True)
        assert ret["changes"] == {"test": None}
        assert envstate.os.environ == {"INITIAL": "initial"}


def test_setenv_clearall():
    """
    test that ``clear_all`` option sets other values to ''
    """
    with patch.dict(os.environ, {"INITIAL": "initial"}, clear=True):
        ret = envstate.setenv("test", "value", clear_all=True)
        assert ret["changes"] == {"test": "value", "INITIAL": ""}
        if salt.utils.platform.is_windows():
            assert envstate.os.environ == {"TEST": "value", "INITIAL": ""}
        else:
            assert envstate.os.environ == {"test": "value", "INITIAL": ""}


def test_setenv_clearall_with_unset():
    """
    test that ``clear_all`` option combined with ``false_unsets``
    unsets other values from environment
    """
    with patch.dict(os.environ, {"INITIAL": "initial"}, clear=True):
        ret = envstate.setenv("test", "value", false_unsets=True, clear_all=True)
        assert ret["changes"] == {"test": "value", "INITIAL": None}
        if salt.utils.platform.is_windows():
            assert envstate.os.environ == {"TEST": "value"}
        else:
            assert envstate.os.environ == {"test": "value"}


def test_setenv_unset_multi():
    """
    test basically same things that above tests but with multiple values passed
    """
    with patch.dict(os.environ, {"INITIAL": "initial"}, clear=True):
        ret = envstate.setenv("notimportant", {"foo": "bar"})
        assert ret["changes"] == {"foo": "bar"}

        with patch.dict(envstate.__utils__, {"reg.read_value": MagicMock()}):
            ret = envstate.setenv(
                "notimportant", {"test": False, "foo": "baz"}, false_unsets=True
            )
        assert ret["changes"] == {"test": None, "foo": "baz"}
        if salt.utils.platform.is_windows():
            assert envstate.os.environ == {"INITIAL": "initial", "FOO": "baz"}
        else:
            assert envstate.os.environ == {"INITIAL": "initial", "foo": "baz"}

        with patch.dict(envstate.__utils__, {"reg.read_value": MagicMock()}):
            ret = envstate.setenv("notimportant", {"test": False, "foo": "bax"})
        assert ret["changes"] == {"test": "", "foo": "bax"}
        if salt.utils.platform.is_windows():
            assert envstate.os.environ == {
                "INITIAL": "initial",
                "FOO": "bax",
                "TEST": "",
            }
        else:
            assert envstate.os.environ == {
                "INITIAL": "initial",
                "foo": "bax",
                "test": "",
            }


def test_setenv_test_mode():
    """
    test that imitating action returns good values
    """
    with patch.dict(os.environ, {"INITIAL": "initial"}, clear=True):
        with patch.dict(envstate.__opts__, {"test": True}):
            ret = envstate.setenv("test", "value")
            assert ret["changes"] == {"test": "value"}
            ret = envstate.setenv("INITIAL", "initial")
            assert ret["changes"] == {}



"""
HTTP Logstash engine
==========================

An engine that reads messages from the salt event bus and pushes
them onto a logstash endpoint via HTTP requests.

.. versionchanged:: 2018.3.0

.. note::
    By default, this engine take everything from the Salt bus and exports into
    Logstash.
    For a better selection of the events that you want to publish, you can use
    the ``tags`` and ``funs`` options.

:configuration: Example configuration

    .. code-block:: yaml

        engines:
          - http_logstash:
              url: http://blabla.com/salt-stuff
              tags:
                  - salt/job/*/new
                  - salt/job/*/ret/*
              funs:
                  - probes.results
                  - bgp.config
"""

import fnmatch

import salt.utils.event
import salt.utils.http
import salt.utils.json

_HEADERS = {"Content-Type": "application/json"}


def _logstash(url, data):
    """
    Issues HTTP queries to the logstash server.
    """
    result = salt.utils.http.query(
        url,
        "POST",
        header_dict=_HEADERS,
        data=salt.utils.json.dumps(data),
        decode=True,
        status=True,
        opts=__opts__,
    )
    return result


def start(url, funs=None, tags=None):
    """
    Listen to salt events and forward them to logstash.

    url
        The Logstash endpoint.

    funs: ``None``
        A list of functions to be compared against, looking into the ``fun``
        field from the event data. This option helps to select the events
        generated by one or more functions.
        If an event does not have the ``fun`` field in the data section, it
        will be published. For a better selection, consider using the ``tags``
        option.
        By default, this option accepts any event to be submitted to Logstash.

    tags: ``None``
        A list of pattern to compare the event tag against.
        By default, this option accepts any event to be submitted to Logstash.
    """
    if __opts__.get("id").endswith("_master"):
        instance = "master"
    else:
        instance = "minion"
    with salt.utils.event.get_event(
        instance,
        sock_dir=__opts__["sock_dir"],
        opts=__opts__,
    ) as event_bus:
        while True:
            event = event_bus.get_event(full=True)
            if event:
                publish = True
                if tags and isinstance(tags, list):
                    found_match = False
                    for tag in tags:
                        if fnmatch.fnmatch(event["tag"], tag):
                            found_match = True
                    publish = found_match
                if funs and "fun" in event["data"]:
                    if not event["data"]["fun"] in funs:
                        publish = False
                if publish:
                    _logstash(url, event["data"])



"""
tests.integration.states.match
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import pytest

from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS


class StateMatchTest(ModuleCase):
    """
    Validate the file state
    """

    @pytest.mark.slow_test
    def test_issue_2167_ipcidr_no_AttributeError(self):
        subnets = self.run_function("network.subnets")
        self.assertTrue(len(subnets) > 0)
        sls_contents = """
        base:
          {}:
            - match: ipcidr
            - test
        """.format(
            subnets[0]
        )
        top_filename = "issue-2167-ipcidr-match.sls"
        with pytest.helpers.temp_file(
            top_filename, sls_contents, RUNTIME_VARS.TMP_BASEENV_STATE_TREE
        ):
            ret = self.run_function("state.top", [top_filename])
            self.assertNotIn(
                "AttributeError: 'Matcher' object has no attribute 'functions'", ret
            )



"""
Functions to work with JSON
"""

import contextlib
import json
import logging

import salt.utils.data
import salt.utils.stringutils

log = logging.getLogger(__name__)


# One to one mappings
JSONEncoder = json.JSONEncoder


def __split(raw):
    """
    Performs a splitlines on the string. This function exists to make mocking
    possible in unit tests, since the member functions of the str/unicode
    builtins cannot be mocked.
    """
    return raw.splitlines()


def find_json(s: str):
    """Pass in a string and load JSON within it.

    The string may contain non-JSON text before and after the JSON document.

    Raises ValueError if no valid JSON was found.
    """
    decoder = json.JSONDecoder()

    # We look for the beginning of JSON objects / arrays and let raw_decode() handle
    # extraneous data at the end.
    for idx, char in enumerate(s):
        if char == "{" or char == "[":
            # JSONDecodeErrors are expected on stray '{'/'[' in the non-JSON part
            with contextlib.suppress(json.JSONDecodeError):
                data, _ = decoder.raw_decode(s[idx:])
                return data

    raise ValueError


def import_json():
    """
    Import a json module, starting with the quick ones and going down the list)
    """
    for fast_json in ("ujson", "yajl", "json"):
        try:
            mod = __import__(fast_json)
            log.trace("loaded %s json lib", fast_json)
            return mod
        except ImportError:
            continue


def load(fp, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Wraps json.load

    You can pass an alternate json module (loaded via import_json() above)
    using the _json_module argument)
    """
    return kwargs.pop("_json_module", json).load(fp, **kwargs)


def loads(s, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Wraps json.loads and prevents a traceback in the event that a bytestring is
    passed to the function. (Python < 3.6 cannot load bytestrings)

    You can pass an alternate json module (loaded via import_json() above)
    using the _json_module argument)
    """
    json_module = kwargs.pop("_json_module", json)
    try:
        return json_module.loads(s, **kwargs)
    except TypeError as exc:
        # json.loads cannot load bytestrings in Python < 3.6
        if isinstance(s, bytes):
            return json_module.loads(salt.utils.stringutils.to_unicode(s), **kwargs)
        else:
            raise


def dump(obj, fp, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Wraps json.dump, and assumes that ensure_ascii is False (unless explicitly
    passed as True) for unicode compatibility. Note that setting it to True
    will mess up any unicode characters, as they will be dumped as the string
    literal version of the unicode code point.

    On Python 2, encodes the result to a str since json.dump does not want
    unicode types.

    You can pass an alternate json module (loaded via import_json() above)
    using the _json_module argument)
    """
    json_module = kwargs.pop("_json_module", json)
    if "ensure_ascii" not in kwargs:
        kwargs["ensure_ascii"] = False
    return json_module.dump(obj, fp, **kwargs)


def dumps(obj, **kwargs):
    """
    .. versionadded:: 2018.3.0

    Wraps json.dumps, and assumes that ensure_ascii is False (unless explicitly
    passed as True) for unicode compatibility. Note that setting it to True
    will mess up any unicode characters, as they will be dumped as the string
    literal version of the unicode code point.

    On Python 2, encodes the result to a str since json.dumps does not want
    unicode types.

    You can pass an alternate json module (loaded via import_json() above)
    using the _json_module argument)
    """
    json_module = kwargs.pop("_json_module", json)
    if "ensure_ascii" not in kwargs:
        kwargs["ensure_ascii"] = False
    return json_module.dumps(obj, **kwargs)



import pytest

import salt.loader
import salt.pillar
import salt.state
import salt.utils.cache
import salt.utils.jinja
from tests.support.mock import patch


@pytest.fixture
def mock_cached_loader():
    """
    Mock the SaltCacheLoader

    The mock keeps track of the number of
    instantiations and the most recent args and kwargs.
    """

    class CacheLoader(salt.utils.jinja.SaltCacheLoader):
        args = []
        kwargs = {}
        calls = 0

        def __init__(self, *args, **kwargs):
            self.__class__.calls += 1
            self.__class__.args = args
            self.__class__.kwargs = kwargs
            super().__init__(*args, **kwargs)

    with patch("salt.utils.jinja.SaltCacheLoader", CacheLoader):
        yield CacheLoader


def test_pillar_tops(temp_salt_master, temp_salt_minion, tmp_path, mock_cached_loader):
    """
    pillar fileclient is used while rendering pillar tops
    """
    tops = """
    base:
      "*":
        - test_pillar
    """
    pillarsls = """
    foo: bar
    """
    opts = temp_salt_master.config.copy()

    with temp_salt_master.pillar_tree.base.temp_file("top.sls", tops):
        with temp_salt_master.pillar_tree.base.temp_file("test_pillar.sls", pillarsls):
            grains = salt.loader.grains(opts)
            pillar = salt.pillar.Pillar(
                opts,
                grains,
                temp_salt_minion.id,
                "base",
            )
            pillar.get_tops()
            assert mock_cached_loader.calls == 1
            assert "_file_client" in mock_cached_loader.kwargs
            assert mock_cached_loader.kwargs["_file_client"] == pillar.client


def test_pillar_render(
    temp_salt_master, temp_salt_minion, tmp_path, mock_cached_loader
):
    """
    pillar fileclient is used while rendering jinja pillar
    """
    tops = """
    base:
      "*":
        - test_pillar
    """
    pillarsls = """
    foo: bar
    """
    opts = temp_salt_master.config.copy()

    with temp_salt_master.pillar_tree.base.temp_file("top.sls", tops):
        with temp_salt_master.pillar_tree.base.temp_file("test_pillar.sls", pillarsls):
            grains = salt.loader.grains(opts)
            pillar = salt.pillar.Pillar(
                opts,
                grains,
                temp_salt_minion.id,
                "base",
            )
            pdata = pillar.render_pillar({"base": ["test_pillar"]})
            assert pdata == ({"foo": "bar"}, [])
            assert mock_cached_loader.calls == 1
            assert "_file_client" in mock_cached_loader.kwargs
            assert mock_cached_loader.kwargs["_file_client"] == pillar.client


def test_highstate(temp_salt_master, temp_salt_minion, tmp_path, mock_cached_loader):
    """
    pillar fileclient is used while rendering pillar tops
    """
    statesls = """
    test_state:
      cmd.run:
        - name: echos foo
    """
    opts = temp_salt_master.config.copy()

    with temp_salt_master.state_tree.base.temp_file("test_state.sls", statesls):
        highstate = salt.state.HighState(
            opts,
        )
        a = highstate.render_highstate({"base": ["test_state"]})
        assert mock_cached_loader.calls == 1
        assert "_file_client" in mock_cached_loader.kwargs
        assert mock_cached_loader.kwargs["_file_client"] == highstate.client



"""
React by calling asynchronous runners
"""

import salt.runner


def cmd(name, func=None, arg=(), **kwargs):
    """
    Execute a runner asynchronous:

    USAGE:

    .. code-block:: yaml

        run_cloud:
          runner.cmd:
            - func: cloud.create
            - arg:
                - my-ec2-config
                - myinstance

        run_cloud:
          runner.cmd:
            - func: cloud.create
            - kwargs:
                provider: my-ec2-config
                instances: myinstance
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if func is None:
        func = name
    local_opts = {}
    local_opts.update(__opts__)
    local_opts["async"] = True  # ensure this will be run asynchronous
    local_opts.update({"fun": func, "arg": arg, "kwarg": kwargs})
    runner = salt.runner.Runner(local_opts)
    runner.run()
    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.debconfmod as debconfmod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {debconfmod: {}}


def test_set_file():
    """
    Test to set debconf selections from a file or a template
    """
    name = "nullmailer"
    source = "salt://pathto/pkg.selections"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    comt = "Context must be formed as a dict"
    ret.update({"comment": comt})
    assert debconfmod.set_file(name, source, context="salt") == ret

    comt = "Defaults must be formed as a dict"
    ret.update({"comment": comt})
    assert debconfmod.set_file(name, source, defaults="salt") == ret

    with patch.dict(debconfmod.__opts__, {"test": True}):
        comt = "Debconf selections would have been set."
        ret.update({"comment": comt, "result": None})
        assert debconfmod.set_file(name, source) == ret

        with patch.dict(debconfmod.__opts__, {"test": False}):
            mock = MagicMock(return_value=True)
            with patch.dict(debconfmod.__salt__, {"debconf.set_file": mock}):
                comt = "Debconf selections were set."
                ret.update({"comment": comt, "result": True})
                assert debconfmod.set_file(name, source) == ret


def test_set():
    """
    Test to set debconf selections
    """
    name = "nullmailer"
    data = {
        "shared/mailname": {"type": "string", "value": "server.domain.tld"},
        "nullmailer/relayhost": {"type": "string", "value": "mail.domain.tld"},
    }

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    changes = {
        "nullmailer/relayhost": "New value: mail.domain.tld",
        "shared/mailname": "New value: server.domain.tld",
    }

    mock = MagicMock(return_value=None)
    with patch.dict(debconfmod.__salt__, {"debconf.show": mock}):
        with patch.dict(debconfmod.__opts__, {"test": True}):
            ret.update({"changes": changes})
            assert debconfmod.set(name, data) == ret



"""
    Test cases for salt.modules.syslog_ng
"""

import os
from textwrap import dedent

import pytest

import salt.modules.syslog_ng as syslog_ng
from tests.support.mock import MagicMock, patch


@pytest.fixture
def _version():
    return "3.6.0alpha0"


@pytest.fixture
def _modules():
    return (
        "syslogformat,json-plugin,basicfuncs,afstomp,afsocket,cryptofuncs,"
        "afmongodb,dbparser,system-source,affile,pseudofile,afamqp,"
        "afsocket-notls,csvparser,linux-kmsg-format,afuser,confgen,afprog"
    )


@pytest.fixture
def version_output(_version, _modules):
    return """syslog-ng {0}
Installer-Version: {0}
Revision:
Compile-Date: Apr  4 2014 20:26:18
Error opening plugin module; module='afsocket-tls', error='/home/tibi/install/syslog-ng/lib/syslog-ng/libafsocket-tls.so: undefined symbol: tls_context_setup_session'
Available-Modules: {1}
Enable-Debug: on
Enable-GProf: off
Enable-Memtrace: off
Enable-IPv6: on
Enable-Spoof-Source: off
Enable-TCP-Wrapper: off
Enable-Linux-Caps: off""".format(
        _version, _modules
    )


@pytest.fixture
def stats_output():
    return """SourceName;SourceId;SourceInstance;State;Type;Number
center;;received;a;processed;0
destination;#anon-destination0;;a;processed;0
destination;#anon-destination1;;a;processed;0
source;s_gsoc2014;;a;processed;0
center;;queued;a;processed;0
global;payload_reallocs;;a;processed;0
global;sdata_updates;;a;processed;0
global;msg_clones;;a;processed;0"""


@pytest.fixture
def orig_env():
    return {"PATH": "/foo:/bar"}


@pytest.fixture
def bin_dir():
    return "/baz"


@pytest.fixture
def mocked_env():
    return {"PATH": "/foo:/bar:/baz"}


@pytest.fixture
def configure_loader_modules():
    return {syslog_ng: {}}


def test_statement_without_options():
    s = syslog_ng.Statement("source", "s_local", options=[])
    b = s.build()
    assert b == (
        dedent(
            """\
        source s_local {
        };
        """
        )
    )


def test_non_empty_statement():
    o1 = syslog_ng.Option("file")
    o2 = syslog_ng.Option("tcp")
    s = syslog_ng.Statement("source", "s_local", options=[o1, o2])
    b = s.build()
    assert b == (
        dedent(
            """\
        source s_local {
            file(
            );
            tcp(
            );
        };
        """
        )
    )


def test_option_with_parameters():
    o1 = syslog_ng.Option("file")
    p1 = syslog_ng.SimpleParameter('"/var/log/messages"')
    p2 = syslog_ng.SimpleParameter()
    p3 = syslog_ng.TypedParameter()
    p3.type = "tls"
    p2.value = '"/var/log/syslog"'
    o1.add_parameter(p1)
    o1.add_parameter(p2)
    o1.add_parameter(p3)
    b = o1.build()
    assert b == (
        dedent(
            """\
        file(
            "/var/log/messages",
            "/var/log/syslog",
            tls(
            )
        );
        """
        )
    )


def test_parameter_with_values():
    p = syslog_ng.TypedParameter()
    p.type = "tls"
    v1 = syslog_ng.TypedParameterValue()
    v1.type = "key_file"

    v2 = syslog_ng.TypedParameterValue()
    v2.type = "cert_file"

    p.add_value(v1)
    p.add_value(v2)

    b = p.build()
    assert b == (
        dedent(
            """\
        tls(
            key_file(
            ),
            cert_file(
            )
        )"""
        )
    )


def test_value_with_arguments():
    t = syslog_ng.TypedParameterValue()
    t.type = "key_file"

    a1 = syslog_ng.Argument('"/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"')
    a2 = syslog_ng.Argument('"/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"')

    t.add_argument(a1)
    t.add_argument(a2)

    b = t.build()
    assert b == (
        dedent(
            """\
        key_file(
            "/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"
            "/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"
        )"""
        )
    )


def test_end_to_end_statement_generation():
    s = syslog_ng.Statement("source", "s_tls")

    o = syslog_ng.Option("tcp")

    ip = syslog_ng.TypedParameter("ip")
    ip.add_value(syslog_ng.SimpleParameterValue("'192.168.42.2'"))
    o.add_parameter(ip)

    port = syslog_ng.TypedParameter("port")
    port.add_value(syslog_ng.SimpleParameterValue(514))
    o.add_parameter(port)

    tls = syslog_ng.TypedParameter("tls")
    key_file = syslog_ng.TypedParameterValue("key_file")
    key_file.add_argument(
        syslog_ng.Argument('"/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"')
    )
    cert_file = syslog_ng.TypedParameterValue("cert_file")
    cert_file.add_argument(
        syslog_ng.Argument('"/opt/syslog-ng/etc/syslog-ng/cert.d/syslog-ng.cert"')
    )
    peer_verify = syslog_ng.TypedParameterValue("peer_verify")
    peer_verify.add_argument(syslog_ng.Argument("optional-untrusted"))
    tls.add_value(key_file)
    tls.add_value(cert_file)
    tls.add_value(peer_verify)
    o.add_parameter(tls)

    s.add_child(o)
    b = s.build()
    assert b == (
        dedent(
            """\
        source s_tls {
            tcp(
                ip(
                    '192.168.42.2'
                ),
                port(
                    514
                ),
                tls(
                    key_file(
                        "/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"
                    ),
                    cert_file(
                        "/opt/syslog-ng/etc/syslog-ng/cert.d/syslog-ng.cert"
                    ),
                    peer_verify(
                        optional-untrusted
                    )
                )
            );
        };
        """
        )
    )


@pytest.mark.skip_on_windows(reason="Module not available on Windows")
def test_version(_version, version_output, orig_env, bin_dir, mocked_env):
    cmd_ret = {"retcode": 0, "stdout": version_output}
    expected_output = {"retcode": 0, "stdout": _version}
    cmd_args = ["syslog-ng", "-V"]

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        result = syslog_ng.version()
        assert result == expected_output
        cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        result = syslog_ng.version(syslog_ng_sbin_dir=bin_dir)
        assert result == expected_output
        cmd_mock.assert_called_once_with(cmd_args, env=mocked_env, python_shell=False)


@pytest.mark.skip_on_windows(reason="Module not available on Windows")
def test_stats(stats_output, orig_env, bin_dir, mocked_env):
    cmd_ret = {"retcode": 0, "stdout": stats_output}
    cmd_args = ["syslog-ng-ctl", "stats"]

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        result = syslog_ng.stats()
        assert result == cmd_ret
        cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        result = syslog_ng.stats(syslog_ng_sbin_dir=bin_dir)
        assert result == cmd_ret
        cmd_mock.assert_called_once_with(cmd_args, env=mocked_env, python_shell=False)


@pytest.mark.skip_on_windows(reason="Module not available on Windows")
def test_modules(_modules, version_output, orig_env, bin_dir, mocked_env):
    cmd_ret = {"retcode": 0, "stdout": version_output}
    expected_output = {"retcode": 0, "stdout": _modules}
    cmd_args = ["syslog-ng", "-V"]

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        result = syslog_ng.modules()
        assert result == expected_output
        cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        result = syslog_ng.modules(syslog_ng_sbin_dir=bin_dir)
        assert result == expected_output
        cmd_mock.assert_called_once_with(cmd_args, env=mocked_env, python_shell=False)


@pytest.mark.skip_on_windows(reason="Module not available on Windows")
def test_config_test(orig_env, bin_dir, mocked_env):
    cmd_ret = {"retcode": 0, "stderr": "", "stdout": "Foo"}
    cmd_args = ["syslog-ng", "--syntax-only"]

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        result = syslog_ng.config_test()
        assert result == cmd_ret
        cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        result = syslog_ng.config_test(syslog_ng_sbin_dir=bin_dir)
        assert result == cmd_ret
        cmd_mock.assert_called_once_with(cmd_args, env=mocked_env, python_shell=False)


@pytest.mark.skip_on_windows(reason="Module not available on Windows")
def test_config_test_cfgfile(orig_env, bin_dir, mocked_env):
    cfgfile = "/path/to/syslog-ng.conf"
    cmd_ret = {"retcode": 1, "stderr": "Syntax error...", "stdout": ""}
    cmd_args = ["syslog-ng", "--syntax-only", f"--cfgfile={cfgfile}"]

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        assert syslog_ng.config_test(cfgfile=cfgfile) == cmd_ret
        cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

    cmd_mock = MagicMock(return_value=cmd_ret)
    with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
        os.environ, orig_env
    ):
        assert (
            syslog_ng.config_test(syslog_ng_sbin_dir=bin_dir, cfgfile=cfgfile)
            == cmd_ret
        )
        cmd_mock.assert_called_once_with(cmd_args, env=mocked_env, python_shell=False)



"""
Support for Bluetooth (using BlueZ in Linux).

The following packages are required packages for this module:

    bluez >= 5.7
    bluez-libs >= 5.7
    bluez-utils >= 5.7
    pybluez >= 0.18
"""

import shlex

import salt.utils.validate.net
from salt.exceptions import CommandExecutionError

HAS_PYBLUEZ = False
try:
    import bluetooth  # pylint: disable=import-error

    HAS_PYBLUEZ = True
except ImportError:
    pass

__func_alias__ = {"address_": "address"}

# Define the module's virtual name
__virtualname__ = "bluetooth"


def __virtual__():
    """
    Only load the module if bluetooth is installed
    """
    if HAS_PYBLUEZ:
        return __virtualname__
    return (
        False,
        "The bluetooth execution module cannot be loaded: bluetooth not installed.",
    )


def version():
    """
    Return Bluez version from bluetoothd -v

    CLI Example:

    .. code-block:: bash

        salt '*' bluetoothd.version
    """
    cmd = "bluetoothctl -v"
    out = __salt__["cmd.run"](cmd).splitlines()
    bluez_version = out[0]
    pybluez_version = "<= 0.18 (Unknown, but installed)"
    try:
        pybluez_version = bluetooth.__version__
    except Exception as exc:  # pylint: disable=broad-except
        pass
    return {"Bluez": bluez_version, "PyBluez": pybluez_version}


def address_():
    """
    Get the many addresses of the Bluetooth adapter

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.address
    """
    ret = {}
    cmd = "hciconfig"
    out = __salt__["cmd.run"](cmd).splitlines()
    dev = ""
    for line in out:
        if line.startswith("hci"):
            comps = line.split(":")
            dev = comps[0]
            ret[dev] = {
                "device": dev,
                "path": f"/sys/class/bluetooth/{dev}",
            }
        if "BD Address" in line:
            comps = line.split()
            ret[dev]["address"] = comps[2]
        if "DOWN" in line:
            ret[dev]["power"] = "off"
        if "UP RUNNING" in line:
            ret[dev]["power"] = "on"
    return ret


def power(dev, mode):
    """
    Power a bluetooth device on or off

    CLI Examples:

    .. code-block:: bash

        salt '*' bluetooth.power hci0 on
        salt '*' bluetooth.power hci0 off
    """
    if dev not in address_():
        raise CommandExecutionError("Invalid dev passed to bluetooth.power")

    if mode == "on" or mode is True:
        state = "up"
        mode = "on"
    else:
        state = "down"
        mode = "off"
    cmd = f"hciconfig {dev} {state}"
    __salt__["cmd.run"](cmd).splitlines()
    info = address_()
    if info[dev]["power"] == mode:
        return True
    return False


def discoverable(dev):
    """
    Enable this bluetooth device to be discoverable.

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.discoverable hci0
    """
    if dev not in address_():
        raise CommandExecutionError("Invalid dev passed to bluetooth.discoverable")

    cmd = f"hciconfig {dev} iscan"
    __salt__["cmd.run"](cmd).splitlines()
    cmd = f"hciconfig {dev}"
    out = __salt__["cmd.run"](cmd)
    if "UP RUNNING ISCAN" in out:
        return True
    return False


def noscan(dev):
    """
    Turn off scanning modes on this device.

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.noscan hci0
    """
    if dev not in address_():
        raise CommandExecutionError("Invalid dev passed to bluetooth.noscan")

    cmd = f"hciconfig {dev} noscan"
    __salt__["cmd.run"](cmd).splitlines()
    cmd = f"hciconfig {dev}"
    out = __salt__["cmd.run"](cmd)
    if "SCAN" in out:
        return False
    return True


def scan():
    """
    Scan for bluetooth devices in the area

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.scan
    """
    ret = []
    devices = bluetooth.discover_devices(lookup_names=True)
    for device in devices:
        ret.append({device[0]: device[1]})
    return ret


def block(bdaddr):
    """
    Block a specific bluetooth device by BD Address

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.block DE:AD:BE:EF:CA:FE
    """
    if not salt.utils.validate.net.mac(bdaddr):
        raise CommandExecutionError("Invalid BD address passed to bluetooth.block")

    cmd = f"hciconfig {bdaddr} block"
    __salt__["cmd.run"](cmd).splitlines()


def unblock(bdaddr):
    """
    Unblock a specific bluetooth device by BD Address

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.unblock DE:AD:BE:EF:CA:FE
    """
    if not salt.utils.validate.net.mac(bdaddr):
        raise CommandExecutionError("Invalid BD address passed to bluetooth.unblock")

    cmd = f"hciconfig {bdaddr} unblock"
    __salt__["cmd.run"](cmd).splitlines()


def pair(address, key):
    """
    Pair the bluetooth adapter with a device

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.pair DE:AD:BE:EF:CA:FE 1234

    Where DE:AD:BE:EF:CA:FE is the address of the device to pair with, and 1234
    is the passphrase.

    TODO: This function is currently broken, as the bluez-simple-agent program
    no longer ships with BlueZ >= 5.0. It needs to be refactored.
    """
    if not salt.utils.validate.net.mac(address):
        raise CommandExecutionError("Invalid BD address passed to bluetooth.pair")

    try:
        int(key)
    except Exception:  # pylint: disable=broad-except
        raise CommandExecutionError(
            "bluetooth.pair requires a numerical key to be used"
        )

    addy = address_()
    cmd = "echo {} | bluez-simple-agent {} {}".format(
        shlex.quote(addy["device"]), shlex.quote(address), shlex.quote(key)
    )
    out = __salt__["cmd.run"](cmd, python_shell=True).splitlines()
    return out


def unpair(address):
    """
    Unpair the bluetooth adapter from a device

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.unpair DE:AD:BE:EF:CA:FE

    Where DE:AD:BE:EF:CA:FE is the address of the device to unpair.

    TODO: This function is currently broken, as the bluez-simple-agent program
    no longer ships with BlueZ >= 5.0. It needs to be refactored.
    """
    if not salt.utils.validate.net.mac(address):
        raise CommandExecutionError("Invalid BD address passed to bluetooth.unpair")

    cmd = f"bluez-test-device remove {address}"
    out = __salt__["cmd.run"](cmd).splitlines()
    return out


def start():
    """
    Start the bluetooth service.

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.start
    """
    out = __salt__["service.start"]("bluetooth")
    return out


def stop():
    """
    Stop the bluetooth service.

    CLI Example:

    .. code-block:: bash

        salt '*' bluetooth.stop
    """
    out = __salt__["service.stop"]("bluetooth")
    return out



"""
Module for returning various status data about a minion. These data can be
useful for compiling into stats later, or for problem-solving if your minion is
having problems.

.. versionadded:: 0.12.0

:depends:  - wmi
"""

import ctypes
import datetime
import logging

import salt.utils.event
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.win_pdh
from salt.modules.status import ping_master, time_
from salt.utils.functools import namespaced_function
from salt.utils.network import host_to_ips as _host_to_ips

log = logging.getLogger(__name__)

try:
    if salt.utils.platform.is_windows():
        import wmi

        import salt.utils.winapi

        HAS_WMI = True
    else:
        HAS_WMI = False
except ImportError:
    HAS_WMI = False

try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False

if salt.utils.platform.is_windows():
    ping_master = namespaced_function(ping_master, globals())
    time_ = namespaced_function(time_, globals())

__virtualname__ = "status"


# Taken from https://www.geoffchappell.com/studies/windows/km/ntoskrnl/api/ex/sysinfo/performance.htm
class SYSTEM_PERFORMANCE_INFORMATION(ctypes.Structure):
    _fields_ = [
        ("IdleProcessTime", ctypes.c_int64),
        ("IoReadTransferCount", ctypes.c_int64),
        ("IoWriteTransferCount", ctypes.c_int64),
        ("IoOtherTransferCount", ctypes.c_int64),
        ("IoReadOperationCount", ctypes.c_ulong),
        ("IoWriteOperationCount", ctypes.c_ulong),
        ("IoOtherOperationCount", ctypes.c_ulong),
        ("AvailablePages", ctypes.c_ulong),
        ("CommittedPages", ctypes.c_ulong),
        ("CommitLimit", ctypes.c_ulong),
        ("PeakCommitment", ctypes.c_ulong),
        ("PageFaultCount", ctypes.c_ulong),
        ("CopyOnWriteCount", ctypes.c_ulong),
        ("TransitionCount", ctypes.c_ulong),
        ("CacheTransitionCount", ctypes.c_ulong),
        ("DemandZeroCount", ctypes.c_ulong),
        ("PageReadCount", ctypes.c_ulong),
        ("PageReadIoCount", ctypes.c_ulong),
        ("CacheReadCount", ctypes.c_ulong),  # Was c_ulong ** 2
        ("CacheIoCount", ctypes.c_ulong),
        ("DirtyPagesWriteCount", ctypes.c_ulong),
        ("DirtyWriteIoCount", ctypes.c_ulong),
        ("MappedPagesWriteCount", ctypes.c_ulong),
        ("MappedWriteIoCount", ctypes.c_ulong),
        ("PagedPoolPages", ctypes.c_ulong),
        ("NonPagedPoolPages", ctypes.c_ulong),
        ("PagedPoolAllocs", ctypes.c_ulong),
        ("PagedPoolFrees", ctypes.c_ulong),
        ("NonPagedPoolAllocs", ctypes.c_ulong),
        ("NonPagedPoolFrees", ctypes.c_ulong),
        ("FreeSystemPtes", ctypes.c_ulong),
        ("ResidentSystemCodePage", ctypes.c_ulong),
        ("TotalSystemDriverPages", ctypes.c_ulong),
        ("TotalSystemCodePages", ctypes.c_ulong),
        ("NonPagedPoolLookasideHits", ctypes.c_ulong),
        ("PagedPoolLookasideHits", ctypes.c_ulong),
        ("AvailablePagedPoolPages", ctypes.c_ulong),
        ("ResidentSystemCachePage", ctypes.c_ulong),
        ("ResidentPagedPoolPage", ctypes.c_ulong),
        ("ResidentSystemDriverPage", ctypes.c_ulong),
        ("CcFastReadNoWait", ctypes.c_ulong),
        ("CcFastReadWait", ctypes.c_ulong),
        ("CcFastReadResourceMiss", ctypes.c_ulong),
        ("CcFastReadNotPossible", ctypes.c_ulong),
        ("CcFastMdlReadNoWait", ctypes.c_ulong),
        ("CcFastMdlReadWait", ctypes.c_ulong),
        ("CcFastMdlReadResourceMiss", ctypes.c_ulong),
        ("CcFastMdlReadNotPossible", ctypes.c_ulong),
        ("CcMapDataNoWait", ctypes.c_ulong),
        ("CcMapDataWait", ctypes.c_ulong),
        ("CcMapDataNoWaitMiss", ctypes.c_ulong),
        ("CcMapDataWaitMiss", ctypes.c_ulong),
        ("CcPinMappedDataCount", ctypes.c_ulong),
        ("CcPinReadNoWait", ctypes.c_ulong),
        ("CcPinReadWait", ctypes.c_ulong),
        ("CcPinReadNoWaitMiss", ctypes.c_ulong),
        ("CcPinReadWaitMiss", ctypes.c_ulong),
        ("CcCopyReadNoWait", ctypes.c_ulong),
        ("CcCopyReadWait", ctypes.c_ulong),
        ("CcCopyReadNoWaitMiss", ctypes.c_ulong),
        ("CcCopyReadWaitMiss", ctypes.c_ulong),
        ("CcMdlReadNoWait", ctypes.c_ulong),
        ("CcMdlReadWait", ctypes.c_ulong),
        ("CcMdlReadNoWaitMiss", ctypes.c_ulong),
        ("CcMdlReadWaitMiss", ctypes.c_ulong),
        ("CcReadAheadIos", ctypes.c_ulong),
        ("CcLazyWriteIos", ctypes.c_ulong),
        ("CcLazyWritePages", ctypes.c_ulong),
        ("CcDataFlushes", ctypes.c_ulong),
        ("CcDataPages", ctypes.c_ulong),
        ("ContextSwitches", ctypes.c_ulong),
        ("FirstLevelTbFills", ctypes.c_ulong),
        ("SecondLevelTbFills", ctypes.c_ulong),
        ("SystemCalls", ctypes.c_ulong),
        # Windows 8 and above
        ("CcTotalDirtyPages", ctypes.c_ulonglong),
        ("CcDirtyPagesThreshold", ctypes.c_ulonglong),
        ("ResidentAvailablePages", ctypes.c_longlong),
        # Windows 10 and above
        ("SharedCommittedPages", ctypes.c_ulonglong),
    ]


def __virtual__():
    """
    Only works on Windows systems with WMI and WinAPI
    """
    if not salt.utils.platform.is_windows():
        return False, "win_status.py: Requires Windows"

    if not HAS_WMI:
        return False, "win_status.py: Requires WMI and WinAPI"

    if not HAS_PSUTIL:
        return False, "win_status.py: Requires psutil"

    return __virtualname__


__func_alias__ = {"time_": "time"}


def cpustats():
    """
    Return information about the CPU.

    Returns
        dict: A dictionary containing information about the CPU stats.

    CLI Example:

    .. code-block:: bash

        salt * status.cpustats
    """
    # Tries to gather information similar to that returned by a Linux machine
    # Avoid using WMI as there's a lot of overhead

    # Time related info
    user, system, idle, interrupt, dpc = psutil.cpu_times()
    cpu = {"user": user, "system": system, "idle": idle, "irq": interrupt, "dpc": dpc}
    # Count related info
    ctx_switches, interrupts, soft_interrupts, sys_calls = psutil.cpu_stats()
    intr = {"irqs": {"irqs": [], "total": interrupts}}
    soft_irq = {"softirqs": [], "total": soft_interrupts}
    return {
        "btime": psutil.boot_time(),
        "cpu": cpu,
        "ctxt": ctx_switches,
        "intr": intr,
        "processes": len(psutil.pids()),
        "softirq": soft_irq,
        "syscalls": sys_calls,
    }


def meminfo():
    """
    Return information about physical and virtual memory on the system.

    Returns:
        dict: A dictionary of information about memory on the system.

    CLI Example:

    .. code-block:: bash

        salt * status.meminfo
    """
    # Get physical memory
    vm_total, vm_available, vm_percent, vm_used, vm_free = psutil.virtual_memory()
    # Get swap memory
    swp_total, swp_used, swp_free, swp_percent, _, _ = psutil.swap_memory()

    def get_unit_value(memory):
        symbols = ("K", "M", "G", "T", "P", "E", "Z", "Y")
        prefix = {}
        for i, s in enumerate(symbols):
            prefix[s] = 1 << (i + 1) * 10
        for s in reversed(symbols):
            if memory >= prefix[s]:
                value = float(memory) / prefix[s]
                return {"unit": s, "value": value}
        return {"unit": "B", "value": memory}

    return {
        "VmallocTotal": get_unit_value(vm_total),
        "VmallocUsed": get_unit_value(vm_used),
        "VmallocFree": get_unit_value(vm_free),
        "VmallocAvail": get_unit_value(vm_available),
        "SwapTotal": get_unit_value(swp_total),
        "SwapUsed": get_unit_value(swp_used),
        "SwapFree": get_unit_value(swp_free),
    }


def vmstats():
    """
    Return information about the virtual memory on the machine.

    Returns:
        dict: A dictionary of virtual memory stats.

    CLI Example:

    .. code-block:: bash

        salt * status.vmstats
    """
    # Setup the SPI Structure
    spi = SYSTEM_PERFORMANCE_INFORMATION()
    retlen = ctypes.c_ulong()

    # 2 means to query System Performance Information and return it in a
    # SYSTEM_PERFORMANCE_INFORMATION Structure
    ctypes.windll.ntdll.NtQuerySystemInformation(
        2, ctypes.byref(spi), ctypes.sizeof(spi), ctypes.byref(retlen)
    )

    # Return each defined field in a dict
    ret = {}
    for field in spi._fields_:
        ret.update({field[0]: getattr(spi, field[0])})

    return ret


def loadavg():
    """
    Returns counter information related to the load of the machine.

    Returns:
        dict: A dictionary of counters.

    CLI Example:

    .. code-block:: bash

        salt * status.loadavg
    """
    # Counter List (obj, instance, counter)
    counter_list = [
        ("Memory", None, "Available Bytes"),
        ("Memory", None, "Pages/sec"),
        ("Paging File", "*", "% Usage"),
        ("Processor", "*", "% Processor Time"),
        ("Processor", "*", "DPCs Queued/sec"),
        ("Processor", "*", "% Privileged Time"),
        ("Processor", "*", "% User Time"),
        ("Processor", "*", "% DPC Time"),
        ("Processor", "*", "% Interrupt Time"),
        ("Server", None, "Work Item Shortages"),
        ("Server Work Queues", "*", "Queue Length"),
        ("System", None, "Processor Queue Length"),
        ("System", None, "Context Switches/sec"),
    ]
    return salt.utils.win_pdh.get_counters(counter_list=counter_list)


def cpuload():
    """
    .. versionadded:: 2015.8.0

    Return the processor load as a percentage.

    CLI Example:

    .. code-block:: bash

       salt '*' status.cpuload
    """
    return psutil.cpu_percent()


def diskusage(human_readable=False, path=None):
    """
    .. versionadded:: 2015.8.0

    Return the disk usage for this minion.

    Args:

        human_readable (:obj:`bool`, optional):
            If ``True``, usage will be in KB/MB/GB etc.

            Default is ``False``.

        path (:obj:`str`, optional):
            The path to the disk. Usually a drive letter such as ``C:\\``.

            Default is ``None``.

    CLI Example:

    .. code-block:: bash

        salt '*' status.diskusage path=c:/salt
    """
    if not path:
        path = "c:/"

    disk_stats = psutil.disk_usage(path)

    total_val = disk_stats.total
    used_val = disk_stats.used
    free_val = disk_stats.free
    percent = disk_stats.percent

    if human_readable:
        total_val = _byte_calc(total_val)
        used_val = _byte_calc(used_val)
        free_val = _byte_calc(free_val)

    return {"total": total_val, "used": used_val, "free": free_val, "percent": percent}


def procs(count=False):
    """
    Return the process data.

    Args:

        count (:obj:`bool`, optional):
            If ``True``, this function will simply return the number of
            processes.

            .. versionadded:: 2015.8.0

            Default is ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' status.procs
        salt '*' status.procs count
    """
    with salt.utils.winapi.Com():
        wmi_obj = wmi.WMI()
        processes = wmi_obj.win32_process()

    # this short circuit's the function to get a short simple proc count.
    if count:
        return len(processes)

    # a propper run of the function, creating a nonsensically long out put.
    process_info = {}
    for proc in processes:
        process_info[proc.ProcessId] = _get_process_info(proc)

    return process_info


def saltmem(human_readable=False):
    """
    .. versionadded:: 2015.8.0

    Returns the amount of memory that salt is using.

    Args:

        human_readable (:obj:`bool`, optional):
            Return the value in a nicely formatted number.

            Default is ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' status.saltmem
        salt '*' status.saltmem human_readable=True
    """
    # psutil.Process defaults to current process (`os.getpid()`)
    p = psutil.Process()

    # Use oneshot to get a snapshot
    with p.oneshot():
        mem = p.memory_info().rss

    if human_readable:
        return _byte_calc(mem)

    return mem


def uptime(human_readable=False):
    """
    .. versionadded:: 2015.8.0

    Return the system uptime for the machine.

    Args:

        human_readable (:obj:`bool`, optional):
            If ``True``, return uptime in human-readable format, otherwise
            return seconds.

            Default is ``False``.

            .. note::
                Human readable format is ``days, hours:min:sec``. Days will only
                be displayed if more than 0

    Returns:
        str: The uptime in seconds or human-readable format depending on the
            value of ``human_readable``

    CLI Example:

    .. code-block:: bash

        salt '*' status.uptime
        salt '*' status.uptime human_readable=True
    """
    # Get startup time
    startup_time = datetime.datetime.fromtimestamp(psutil.boot_time())

    # Subtract startup time from current time to get the uptime of the system
    uptime = datetime.datetime.now() - startup_time

    return str(uptime) if human_readable else uptime.total_seconds()


def _get_process_info(proc):
    """
    Return  process information
    """
    cmd = salt.utils.stringutils.to_unicode(proc.CommandLine or "")
    name = salt.utils.stringutils.to_unicode(proc.Name)
    info = dict(cmd=cmd, name=name, **_get_process_owner(proc))
    return info


def _get_process_owner(process):
    owner = {}
    domain, error_code, user = None, None, None
    try:
        domain, error_code, user = process.GetOwner()
        owner["user"] = salt.utils.stringutils.to_unicode(user)
        owner["user_domain"] = salt.utils.stringutils.to_unicode(domain)
    except Exception as exc:  # pylint: disable=broad-except
        pass
    if not error_code and all((user, domain)):
        owner["user"] = salt.utils.stringutils.to_unicode(user)
        owner["user_domain"] = salt.utils.stringutils.to_unicode(domain)
    elif process.ProcessId in [0, 4] and error_code == 2:
        # Access Denied for System Idle Process and System
        owner["user"] = "SYSTEM"
        owner["user_domain"] = "NT AUTHORITY"
    else:
        log.warning(
            "Error getting owner of process; PID='%s'; Error: %s",
            process.ProcessId,
            error_code,
        )
    return owner


def _byte_calc(val):
    if val < 1024:
        tstr = str(val) + "B"
    elif val < 1038336:
        tstr = str(val / 1024) + "KB"
    elif val < 1073741824:
        tstr = str(val / 1038336) + "MB"
    elif val < 1099511627776:
        tstr = str(val / 1073741824) + "GB"
    else:
        tstr = str(val / 1099511627776) + "TB"
    return tstr


def _get_connected_ips(port):
    """
    List all connections on the system that have an established connection on
    the passed port. This uses psutil.net_connections instead of netstat to be
    locale agnostic.
    """
    connected_ips = set()
    # Let's use psutil to be non-locale specific
    conns = psutil.net_connections()

    for conn in conns:
        if conn.status == psutil.CONN_ESTABLISHED:
            if conn.raddr.port == port:
                log.debug(
                    "%s %s:%s --> %s:%s",
                    conn.status,
                    conn.laddr.ip,
                    conn.laddr.port,
                    conn.raddr.ip,
                    conn.raddr.port,
                )
                connected_ips.add(conn.raddr.ip)

    return connected_ips


def master(master=None, connected=True):
    """
    .. versionadded:: 2015.5.0

    Fire an event if the minion gets disconnected from its master. This
    function is meant to be run via a scheduled job from the minion. If
    master_ip is an FQDN/Hostname, is must be resolvable to a valid IPv4
    address.

    Args:

        master (:obj:`str`, optional):
            The master address, FQDN, or hostname to connect to.

            Default is ``None``.

        connected (:obj:`bool`, optional):
            If ``True``, fire an event if the master is connected, otherwise
            fire an event if the master is disconnected.

            Default is ``True``.

    CLI Example:

    .. code-block:: bash

        salt '*' status.master
    """
    # the default publishing port
    port = 4505
    master_ips = None

    if master:
        master_ips = _host_to_ips(master)

    if not master_ips:
        return

    if __salt__["config.get"]("publish_port") != "":
        port = int(__salt__["config.get"]("publish_port"))

    master_connection_status = False
    connected_ips = _get_connected_ips(port)

    # Get connection status for master
    for master_ip in master_ips:
        if master_ip in connected_ips:
            master_connection_status = True
            break

    # Connection to master is not as expected
    if master_connection_status is not connected:
        with salt.utils.event.get_event(
            "minion", opts=__opts__, listen=False
        ) as event_bus:
            if master_connection_status:
                event_bus.fire_event(
                    {"master": master}, salt.minion.master_event(type="connected")
                )
            else:
                event_bus.fire_event(
                    {"master": master}, salt.minion.master_event(type="disconnected")
                )

    return master_connection_status
